@@ -1,26 +1,198 @@
-# Copyright (c) 2020-2021 The Center for Theoretical Biological Physics (CTBP) - Rice University
+# Copyright (c) 2020-2022 The Center for Theoretical Biological Physics (CTBP) - Rice University
 # This file is from the Open-MiChroM project, released under the MIT License.
 
 R"""  
 The :class:`~.Optimization` classes perform the energy function parameters training of the chromosomes based on experimental Hi-C data.
 """
 
-from simtk.openmm.app import *
-import simtk.openmm as openmm
-import simtk.unit as units
-from sys import stdout, argv
+# with OpenMM 7.7.0, the import calls have changed. So, try both, if needed
+try:
+    try:
+        # >=7.7.0
+        from openmm.app import *
+    except:
+        # earlier
+        print('Unable to load OpenMM as \'openmm\'. Will try the older way \'simtk.openmm\'')
+        from simtk.openmm.app import *
+except:
+    print('Failed to load OpenMM. Check your configuration.')
+
+
 import numpy as np
-from six import string_types
-import os
-import time
 import random
-import h5py
 from scipy.spatial import distance
 import scipy as sp
 import itertools
 from scipy.stats.stats import pearsonr
 from sklearn.preprocessing import normalize
+import os
+import pandas as pd
 
+class AdamTraining:
+    R"""
+    The :class:`~.AdamTraining` class performs the parameters training for each selected loci pair interaction. 
+    
+    Details about the methodology are decribed in "Zhang, Bin, and Peter G. Wolynes. "Topology, structures, and energy landscapes of human chromosomes." Proceedings of the National Academy of Sciences 112.19 (2015): 6062-6067."
+    
+    
+    The :class:`~.AdamTraining` class receive a Hi-C matrix (text file) as input. The parameters :math:`\mu` (mu) and rc are part of the probability of crosslink function :math:`f(r_{i,j}) = \frac{1}{2}\left( 1 + tanh\left[\mu(r_c - r_{i,j}\right] \right)`, where :math:`r_{i,j}` is the spatial distance between loci (beads) *i* and *j*.
+    
+    Args:
+        mu (float, required):
+            Parameter in the probability of crosslink function. (Default value = 2.0).
+        rc (float, required):
+            Parameter in the probability of crosslink function. (Default value = 2.0).
+        eta (float, required):
+            Learning rate applied in each step (Default value = 0.01).
+        beta1 (float, required):
+            The hyper-parameter of Adam are initial decay rates used when estimating the first and second moments of the gradient. (Default value = 0.9).
+        beta2 (float, required):
+            The hyper-parameter of Adam are initial decay rates used when estimating the first and second moments of the gradient. (Default value = 0.999).
+        it (int, required)
+            The iteration step      
+        
+    """
+    def __init__(self, mu=2.0, rc = 2.0, eta=0.01, beta1=0.9, beta2=0.999, epsilon=1e-8, it=1):
+        self.m_dw, self.v_dw = 0, 0
+        self.m_db, self.v_db = 1, 1
+        self.beta1 = beta1
+        self.beta2 = beta2
+        self.epsilon = epsilon
+        self.eta = eta
+        self.mu = mu
+        self.rc = rc
+        self.NFrames = 0
+        self.t = it
+
+    def _update(self, w, b, dw, db):
+        R"""Adam optimization step. This function updates weights and biases for each step.
+        """
+
+        ## dw, db are from current minibatch
+        ## momentum beta 1
+        # *** weights *** #
+        self.m_dw = self.beta1*self.m_dw + (1-self.beta1)*dw
+        # *** biases *** #
+        self.m_db = self.beta1*self.m_db + (1-self.beta1)*db
+
+        ## rms beta 2
+        # *** weights *** #
+        self.v_dw = self.beta2*self.v_dw + (1-self.beta2)*(dw**2)
+        # *** biases *** #
+        self.v_db = self.beta2*self.v_db + (1-self.beta2)*(db)
+
+        ## bias correction
+        m_dw_corr = self.m_dw/(1-self.beta1**self.t)
+        m_db_corr = self.m_db/(1-self.beta1**self.t)
+        v_dw_corr = self.v_dw/(1-self.beta2**self.t)
+        v_db_corr = self.v_db/(1-self.beta2**self.t)
+
+        ## update weights and biases
+        w = w - self.eta*(m_dw_corr/(np.sqrt(v_dw_corr)+self.epsilon))
+        b = b - self.eta*(m_db_corr/(np.sqrt(v_db_corr)+self.epsilon))
+        self.t += 1
+        return w, b
+
+    def getPars(self, HiC, centerRemove=False, centrange=[0,0], cutoff= 0.0):
+        R"""
+        Receives the experimental Hi-C map (Full dense matrix) in a text format and performs the data normalization from Hi-C frequency/counts/reads to probability.
+        
+        Args:
+            HiC (file, required):
+                Experimental Hi-C map (Full dense matrix) in a text format.
+            centerRemove (bool, optional):
+                Whether to set the contact probability of the centromeric region to zero. (Default value: :code:`False`).
+            centrange (list, required if **centerRemove** = :code:`True`)):
+                Range of the centromeric region, *i.e.*, :code:`centrange=[i,j]`, where *i* and *j*  are the initial and final beads in the centromere. (Default value = :code:`[0,0]`).
+            cutoff (float, optional):
+                Cutoff value for reducing the noise in the original data. Values lower than the **cutoff** are considered :math:`0.0`.
+        """
+        allmap = np.loadtxt(HiC)
+
+        r=np.triu(allmap, k=1)
+        r[np.isinf(r)]= 0.0
+        r[np.isnan(r)]= 0.0
+        r = normalize(r, axis=1, norm='max')
+
+        for i in range(len(r)-1):
+            maxElem = r[i][i+1]
+            if (maxElem != np.max(r[i])):
+                for j in range(len(r[i])):
+                    if maxElem != 0.0:
+                        r[i][j] = float(r[i][j] / maxElem)
+                    else:
+                        r[i][j] = 0.0 
+                    if r[i][j] > 1.0:
+                        r[i][j] = 0.5
+
+        rd = np.transpose(r) 
+        self.expHiC = r+rd + np.diag(np.ones(len(r)))
+        
+        if (centerRemove):
+            centrome = range(centrange[0],centrange[1])
+            self.expHiC[centrome,:] = 0.0
+            self.expHiC[:,centrome] = 0.0
+        
+        #remove noise by cutoff    
+        self.expHiC[self.expHiC<cutoff] = 0.0
+
+        self.mask = self.expHiC == 0.0
+
+        self.phi_exp = self.expHiC #np.triu(self.expHiC, k=1)
+        self.Pi = np.zeros(self.phi_exp.shape)
+
+    def probCalc(self, state):
+        R"""
+        Calculates the contact probability matrix for a given state.
+        """
+
+        Pi = 0.5*(1.0 + np.tanh(self.mu*(self.rc - distance.cdist(state,state, 'euclidean'))))
+    
+        self.Pi += Pi
+        self.NFrames += 1
+
+    def _getGrad(self):
+        R"""
+        Calcultes the gradient function.
+        """
+        return (-self.phi_sim + self.phi_exp)
+
+    def getLamb(self, Lambdas, fixedPoints=None):
+        R"""
+        Calculates the Lagrange multipliers of each pair of interaction and returns the matrix containing the energy values for the optimization step.
+        
+        Args:
+            Lambdas (file, required):
+                The matrix containing the energies values used to make the simulation in that step. 
+            fixedPoints (list, optional):
+                List of all pairs (i,j) of interactions that will remain unchanged throughout the optimization procedure.
+        
+        Returns:
+            :math:`(N,N)` :class:`numpy.ndarray`:
+                Returns an updated matrix of interactions between each pair of bead.
+        """
+        self.phi_sim = self.Pi/self.NFrames
+        self.phi_sim[self.mask] = 0.0
+
+        grad = self._getGrad()
+
+        self.lambdas = pd.read_csv(Lambdas, sep=None, engine='python')
+        newlamb_values, newbias_values = self._update(self.lambdas.values, self.lambdas.values, grad, grad)
+
+        self.bias = newbias_values
+
+        if fixedPoints == None:
+            lamb  = pd.DataFrame(newlamb_values,columns=list(self.lambdas.columns.values))
+        else:
+            for p in fixedPoints: #fixedPoints is a list of tuples for iteraction fixed i,j
+                self.mask[p] = True
+
+            lambs_final = np.where(self.mask,self.lambdas.values, newlamb_values)
+            lamb  = pd.DataFrame(lambs_final,columns=list(self.lambdas.columns.values))
+
+        self.error = np.sum(np.absolute(np.triu(self.phi_sim, k=3) - np.triu(self.phi_exp, k=3)))/np.sum(np.triu(self.phi_exp, k=3))
+
+        return (lamb)
 
 class FullTraining:
     R"""
@@ -49,15 +221,14 @@ class FullTraining:
         c_l (float, required if **reduce** = :code:`True`)):
             The the low-resolution cutoff. (Default value = 0.02).
     """
-    def __init__(self, state, expHiC, mu=2.0, rc=2.5, 
+    def __init__(self, expHiC, mu=2.0, rc=2.5, 
                  cutoff=0.0, reduce=True,
-                 pair_h=2, c_h=0.1, pair_l=4, c_l=0.02, gpu=False 
+                 pair_h=2, c_h=0.1, pair_l=4, c_l=0.02
                 ):
             
         self.mi = mu
         self.rc = rc
         self.cutoff = cutoff
-        self.gpu = gpu
         
         self.getHiCexp(expHiC, centerRemove=False, centrange=[0,0])
         self.hic_sparse = sp.sparse.csr_matrix(np.triu(self.expHiC, k=2))
@@ -66,18 +237,21 @@ class FullTraining:
        
 
         self.ind = self.get_indices(self.hic_sparse)
-        if (self.gpu):
-            import pycuda.driver as drv
-            import pycuda.gpuarray as gpuarray
-            import pycuda.autoinit
-            import skcuda.linalg as linalg
 
         self.size = len(self.ind)     
         self.Pi = np.zeros(self.size)
-        self.Prod_dist = np.zeros(self.hic_sparse.shape)
+        self.Prob_dist = np.zeros(self.hic_sparse.shape)
         self.PiPj = np.zeros((self.size,self.size))
         self.NFrames = 0
 
+    def createInitialLambda(self, sequenceFile, outputPath=".", initialGuess=0.0, baseLine=-0.2):
+        lambdas = np.zeros(self.hic_sparse.shape) + baseLine
+        for i in self.ind:
+            lambdas[i] = initialGuess
+        lambdas = np.triu(lambdas) + np.triu(lambdas).T
+
+        self.saveLambdas(sequenceFile=sequenceFile, data=lambdas, outputPath=outputPath, name="lambda_0")
+
     def appCutoff(self, pair_h, c_h, pair_l, c_l):
         R"""
         Applies the cutoff for low- and high-resolution values.
@@ -95,7 +269,8 @@ class FullTraining:
             if (hic_full[i] > c_h):
                 hic_final[i] = hic_full[i]
 
-        print('Non-zero interactions after high-resolution cutoff ({}): {}'.format( c_h, sp.sparse.csr_matrix(np.triu(hic_final, k=2)).nnz ))
+        high_cut_number =   sp.sparse.csr_matrix(np.triu(hic_final, k=2)).nnz
+        print('Non-zero interactions after high-resolution cutoff ({}): {}'.format( c_h, high_cut_number ))
 
         values = [n for n in range(1,N,pair_l)]
         index =  [x for x in itertools.combinations_with_replacement(values, r=2)]
@@ -103,7 +278,8 @@ class FullTraining:
             if (hic_full[i] > c_l):
                 hic_final[i] = hic_full[i]
         self.hic_sparse = sp.sparse.csr_matrix(np.triu(hic_final, k=2))
-        print('Non-zero interactions after low-resolution cutoff ({}): {}'.format(c_l,  self.hic_sparse.nnz ))
+        print('Non-zero interactions after low-resolution cutoff ({}): {}'.format(c_l,  (self.hic_sparse.nnz-high_cut_number) ))
+        print('Total Non-zero interactions: {}'.format(self.hic_sparse.nnz))
 
     
     def get_indices(self, hic):
@@ -161,7 +337,7 @@ class FullTraining:
         self.PiPj += PiPj                                       
 
      
-        self.Prod_dist += Prob
+        self.Prob_dist += Prob
         self.Pi += Pi
         self.NFrames += 1
         
@@ -197,6 +373,13 @@ class FullTraining:
 
 
         return(pearsonr(a1,a2)[0])
+    def saveLambdas(self, sequenceFile, data, outputPath, name):
+        seq = np.loadtxt(sequenceFile, dtype=str)[:,1]
+
+        lamb = pd.DataFrame(data,columns=seq)
+        lamb.to_csv(os.path.join(outputPath, name), index=False)
+        print("{} file save in {}".format(name, outputPath))
+
         
     def getLambdas(self):
         R"""
@@ -301,13 +484,13 @@ class CustomMiChroMTraining:
         R"""
         Calculates the contact probability matrix and the cross term of the Hessian for the Ideal Chromosome optimization.
         """
-        PiPj = np.zeros((dmax,dmax))
         self.Pold += self.P
         self.P = 0.5*(1.0 + np.tanh(self.mu*(self.r_cut - distance.cdist(state,state, 'euclidean'))))
         self.P[self.P<self.cutoff] = 0.0
-        dmaxl = range(dmax)
-        for i, j in itertools.product(dmaxl,dmaxl):
-            PiPj[i,j] = np.mean(np.diagonal(self.P, offset=(i+self.dinit)))*np.mean(np.diagonal(self.P, offset=(j+self.dinit)))
+        Pi = []
+        for i in range(dmax):
+             Pi.append(np.mean(np.diagonal(self.P, offset=(i+self.dinit))))
+        PiPj = np.outer(Pi, Pi)
       
         self.Bij += PiPj
         self.Nframes += 1 
@@ -415,13 +598,7 @@ class CustomMiChroMTraining:
         
         vec = []
         for pcount,q in enumerate(itertools.combinations_with_replacement(just.keys(), r=2)):
-            nt=0
-            for i, j in itertools.product(just[q[0]],just[q[1]]):
-
-                p_instant[ind[0][pcount], ind[1][pcount]] += self.P[i,j]
-
-                nt += 1
-            p_instant[ind[0][pcount], ind[1][pcount]] = p_instant[ind[0][pcount], ind[1][pcount]]/nt #pi
+            p_instant[ind[0][pcount], ind[1][pcount]] = np.average(self.P[np.ix_(just[q[0]], just[q[1]])])
             vec.append(p_instant[ind[0][pcount], ind[1][pcount]])
         vec = np.array(vec)
 
@@ -551,4 +728,4 @@ class CustomMiChroMTraining:
         new[inds] = lambdas
         new[(inds[1], inds[0])] = lambdas 
         
-        return(new)
\ No newline at end of file
+        return(new)
